@@ -23,27 +23,7 @@
 #include "filter/filter.h"
 #include "filter/f-inst.h"
 
-/* Pool for protocol persistent data set by protocols themselves */
-struct proto_common_private {
-  DOMAIN(attrs) lock;
-  struct proto_common_private **locked_at;
-
-  pool *pool;
-};
-
-typedef union proto_common {
-  DOMAIN(attrs) lock;
-  struct proto_common_private priv;
-} proto_common;
-static proto_common proto_common_pub;
-
-LOBJ_UNLOCK_CLEANUP(proto_common, attrs);
-#define PROTO_COMMON_LOCK(pcm)	LOBJ_LOCK(&proto_common_pub, pcm, proto_common, attrs);
-
-/* Top pool for protocol data */
 pool *proto_pool;
-
-
 static TLIST_LIST(proto) global_proto_list;
 
 static list STATIC_LIST_INIT(protocol_list);
@@ -75,22 +55,6 @@ static inline void channel_reimport(struct channel *c, struct rt_feeding_request
   ev_send(proto_event_list(c->proto), &c->reimport_event);
 }
 
-static inline bool channel_reload(struct channel *c, struct rt_feeding_request *rfr)
-{
-  if ((c->in_keep & RIK_PREFILTER) == RIK_PREFILTER)
-  {
-    channel_reimport(c, rfr);
-    return true;
-  }
-  else
-    return c->proto->reload_routes(c, rfr);
-}
-
-static inline void channel_reload_roa(struct channel *c, struct rt_feeding_request *rfr)
-{
-  ASSERT_DIE(channel_reload(c, rfr));
-}
-
 static inline void channel_refeed(struct channel *c, struct rt_feeding_request *rfr)
 {
   CALL(c->proto->refeed_begin, c, rfr);
@@ -103,18 +67,10 @@ static inline int proto_is_done(struct proto *p)
 static inline int channel_is_active(struct channel *c)
 { return (c->channel_state != CS_DOWN); }
 
-static inline enum channel_reloadable channel_reloadable(struct channel *c)
+static inline int channel_reloadable(struct channel *c)
 {
-  /* Import table always reloads locally */
-  if ((c->in_keep & RIK_PREFILTER) == RIK_PREFILTER)
-    return CHANNEL_RELOADABLE_LOCALLY;
-
-  /* The protocol should indicate how the reload actually works */
-  if (c->proto->reload_routes)
-    return c->reloadable;
-
-  /* No reload possible */
-  return CHANNEL_RELOADABLE_NEVER;
+  return c->reloadable && c->proto->reload_routes
+      || ((c->in_keep & RIK_PREFILTER) == RIK_PREFILTER);
 }
 
 static inline void
@@ -276,7 +232,7 @@ proto_add_channel(struct proto *p, struct channel_config *cf)
 
   c->channel_state = CS_DOWN;
   c->last_state_change = current_time();
-  c->reloadable = CHANNEL_RELOADABLE_LOCALLY;
+  c->reloadable = 1;
 
   init_list(&c->roa_subscriptions);
 
@@ -498,7 +454,6 @@ struct roa_subscription {
   void (*refeed_hook)(struct channel *, struct rt_feeding_request *);
   struct lfjour_recipient digest_recipient;
   event update_event;
-  struct settle aspa_settle;
 };
 
 struct roa_reload_request {
@@ -557,33 +512,9 @@ channel_roa_changed(void *_s)
 	  s->c->proto->name, s->c->name);
 }
 
-static void
-channel_aspa_settle_hook(struct settle *set)
-{
-  SKIP_BACK_DECLARE(struct roa_subscription, s, aspa_settle, set);
-  s->refeed_hook(s->c, NULL);
-}
-
-static void
-channel_aspa_changed(void *_s)
-{
-  struct roa_subscription *s = _s;
-
-  bool seen = false;
-  for (
-      struct lfjour_item *it;
-      it = lfjour_get(&s->digest_recipient);
-      lfjour_release(&s->digest_recipient, it)
-      )
-    seen = true;
-
-  if (seen)
-    settle_kick(&s->aspa_settle, s->c->proto->loop);
-}
-
 static inline void (*channel_roa_reload_hook(int dir))(struct channel *, struct rt_feeding_request *)
 {
-  return dir ? channel_reload_roa : channel_refeed;
+  return dir ? channel_reimport : channel_refeed;
 }
 
 static int
@@ -605,64 +536,31 @@ channel_roa_subscribe(struct channel *c, rtable *tab, int dir)
   if (channel_roa_is_subscribed(c, tab, dir))
     return;
 
-  /* The ROA auxiliary table has the base type */
-  ASSERT_DIE((tab->addr_type == NET_IP4) || (tab->addr_type == NET_IP6));
+  rtable *aux = tab->config->roa_aux_table->table;
 
   struct roa_subscription *s = mb_allocz(c->proto->pool, sizeof(struct roa_subscription));
   *s = (struct roa_subscription) {
     .c = c,
-    .tab = tab,
+    .tab = aux,
     .refeed_hook = channel_roa_reload_hook(dir),
     .digest_recipient = {
       .target = proto_work_list(c->proto),
       .event = &s->update_event,
     },
     .update_event = {
-      .data = s,
       .hook = channel_roa_changed,
+      .data = s,
     },
   };
 
   add_tail(&c->roa_subscriptions, &s->roa_node);
 
-  RT_LOCK(tab, t);
+  RT_LOCK(aux, t);
   rt_lock_table(t);
   rt_setup_digestor(t);
   lfjour_register(&t->export_digest->digest, &s->digest_recipient);
 }
 
-static void
-channel_aspa_subscribe(struct channel *c, rtable *tab, int dir)
-{
-  if (channel_roa_is_subscribed(c, tab, dir))
-    return;
-
-  ASSERT_DIE(tab->addr_type == NET_ASPA);
-
-  struct settle_config sc = { .min = 300 MS, .max = 10 S, };
-  struct roa_subscription *s = mb_allocz(c->proto->pool, sizeof(struct roa_subscription));
-  *s = (struct roa_subscription) {
-    .c = c,
-    .tab = tab,
-    .refeed_hook = channel_roa_reload_hook(dir),
-    .digest_recipient = {
-      .target = proto_work_list(c->proto),
-      .event = &s->update_event,
-    },
-    .update_event = {
-      .data = s,
-      .hook = channel_aspa_changed,
-    },
-    .aspa_settle = SETTLE_INIT(&sc, channel_aspa_settle_hook, NULL),
-  };
-
-  add_tail(&c->roa_subscriptions, &s->roa_node);
-
-  RT_LOCK(tab, t);
-  rt_lock_table(t);
-  lfjour_register(&t->export_all.journal, &s->digest_recipient);
-}
-
 static void
 channel_roa_unsubscribe(struct roa_subscription *s)
 {
@@ -673,24 +571,23 @@ channel_roa_unsubscribe(struct roa_subscription *s)
   }
 
   ev_postpone(&s->update_event);
-  settle_cancel(&s->aspa_settle);
 
   rem_node(&s->roa_node);
   mb_free(s);
 }
 
-
 static void
 channel_roa_subscribe_filter(struct channel *c, int dir)
 {
   const struct filter *f = dir ? c->in_filter : c->out_filter;
+  rtable *tab;
   int valid = 1, found = 0;
 
   if ((f == FILTER_ACCEPT) || (f == FILTER_REJECT))
     return;
 
   /* No automatic reload for non-reloadable channels */
-  if (dir && (channel_reloadable(c) != CHANNEL_RELOADABLE_LOCALLY))
+  if (dir && !channel_reloadable(c))
     valid = 0;
 
   struct filter_iterator fit;
@@ -701,17 +598,16 @@ channel_roa_subscribe_filter(struct channel *c, int dir)
     switch (fi->fi_code)
     {
     case FI_ROA_CHECK:
-      if (valid)
-	channel_roa_subscribe(c, fi->i_FI_ROA_CHECK.rtc->roa_aux_table->table, dir);
+      tab = fi->i_FI_ROA_CHECK.rtc->table;
+      if (valid) channel_roa_subscribe(c, tab, dir);
       found = 1;
       break;
 
     case FI_ASPA_CHECK_EXPLICIT:
-      if (valid)
-	channel_aspa_subscribe(c, fi->i_FI_ASPA_CHECK_EXPLICIT.rtc->table, dir);
+      tab = fi->i_FI_ASPA_CHECK_EXPLICIT.rtc->table;
+      if (valid) channel_roa_subscribe(c, tab, dir);
       found = 1;
       break;
-
     default:
       break;
     }
@@ -1144,11 +1040,10 @@ channel_request_reload(struct channel *c, struct rt_feeding_request *cir)
   else
     CD(c, "Full import reload requested");
 
-  if (!channel_reload(c, cir))
-    if (cir && cir->prefilter.mode)
-      cli_msg(-15, "%s.%s: partial reload refused, please run full reload instead", c->proto->name, c->name);
-    else
-      bug("%s.%s: full reload refused");
+  if ((c->in_keep & RIK_PREFILTER) == RIK_PREFILTER)
+    channel_reimport(c, cir);
+  else if (! c->proto->reload_routes(c, cir))
+    cli_msg(-15, "%s.%s: partial reload refused, please run full reload instead", c->proto->name, c->name);
 }
 
 const struct channel_class channel_basic = {
@@ -1510,7 +1405,9 @@ proto_start(struct proto *p)
 
   if (p->cf->loop_order != DOMAIN_ORDER(the_bird))
   {
-    p->loop = birdloop_new(proto_pool, p->cf->loop_order, p->cf->loop_max_latency, "Protocol %s", p->cf->name);
+    p->loop = birdloop_new(proto_pool, p->cf->loop_order,
+	p->cf->thread_group ? p->cf->thread_group->group : NULL,
+	"Protocol %s", p->cf->name);
     p->pool = birdloop_pool(p->loop);
   }
   else
@@ -1563,6 +1460,9 @@ proto_config_new(struct protocol *pr, int class)
   cf->mrtdump = new_config->proto_default_mrtdump;
   cf->loop_order = DOMAIN_ORDER(the_bird);
 
+  thread_group_finalize_config();
+  cf->thread_group = new_config->default_thread_group;
+
   cf->restart_limit = 5 S;
 
   init_list(&cf->channels);
@@ -1696,6 +1596,9 @@ proto_reconfigure(struct proto *p, struct proto_config *oc, struct proto_config
   PD(p, "Reconfigured");
   p->cf = nc;
 
+  if (p->loop != &main_birdloop)
+    birdloop_transfer(p->loop, oc->thread_group->group, nc->thread_group->group);
+
   return 1;
 }
 
@@ -1892,7 +1795,7 @@ protos_do_commit(struct config *new, struct config *old, int type)
     protos_commit_request.phase--;
 
   /* If something is pending, the next round will be called asynchronously from proto_rethink_goal(). */
-  if (new->shutdown || !proto_rethink_goal_pending)
+  if (!proto_rethink_goal_pending)
     protos_do_commit(new, old, type);
 }
 
@@ -1932,10 +1835,7 @@ proto_rethink_goal(struct proto *p)
 
     proto_rem_node(&global_proto_list, p);
     rfree(p->event);
-    {
-      PROTO_COMMON_LOCK(pcm);
-      mb_free(p->message);
-    }
+    mb_free(p->message);
     mb_free(p);
     if (!nc)
       goto done;
@@ -1953,7 +1853,7 @@ proto_rethink_goal(struct proto *p)
     proto_start(p);
 
 done:
-  if (goal_pending && !--proto_rethink_goal_pending && protos_commit_request.new)
+  if (goal_pending && !--proto_rethink_goal_pending)
     protos_do_commit(
 	protos_commit_request.new,
 	protos_commit_request.old,
@@ -1961,18 +1861,6 @@ done:
 	);
 }
 
-struct proto_rethink_goal_deferred {
-  struct deferred_call dc;
-  struct proto *p;
-};
-
-static void
-proto_rethink_goal_deferred(struct deferred_call *dc)
-{
-  SKIP_BACK_DECLARE(struct proto_rethink_goal_deferred, prgd, dc, dc);
-  proto_rethink_goal(prgd->p);
-}
-
 struct proto *
 proto_spawn(struct proto_config *cf, uint disabled)
 {
@@ -1998,11 +1886,7 @@ proto_enable(struct proto *p)
   ASSERT_DIE(birdloop_inside(&main_birdloop));
   bool changed = p->disabled;
   p->disabled = 0;
-  struct proto_rethink_goal_deferred prgd = {
-    .dc.hook = proto_rethink_goal_deferred,
-    .p = p,
-  };
-  defer_call(&prgd.dc, sizeof prgd);
+  proto_rethink_goal(p);
   return changed;
 }
 
@@ -2282,11 +2166,6 @@ void
 protos_build(void)
 {
   proto_pool = rp_new(&root_pool, the_bird_domain.the_bird, "Protocols");
-  proto_common_pub.lock = DOMAIN_NEW(attrs);
-  {
-    PROTO_COMMON_LOCK(pcm);
-    pcm->pool = rp_new(proto_pool, pcm->lock.attrs, "Common persistent protocol data");
-  }
 
   /* Protocol attributes */
   ea_register_init(&ea_name);
@@ -2327,7 +2206,7 @@ protos_build(void)
   struct settle_config cf = {.min = 0, .max = 0};
   proto_state_table_pub.journal.item_done = proto_journal_item_cleanup;
   proto_state_table_pub.journal.item_size = sizeof(struct proto_pending_update);
-  proto_state_table_pub.journal.loop = birdloop_new(&root_pool, DOMAIN_ORDER(service), 1, "proto journal loop");
+  proto_state_table_pub.journal.loop = birdloop_new(&root_pool, DOMAIN_ORDER(service), NULL, "proto journal loop");
   proto_state_table_pub.journal.domain = proto_state_table_pub.lock.rtable;
 
   lfjour_init(&proto_state_table_pub.journal, &cf);
@@ -2397,8 +2276,6 @@ proto_schedule_down(struct proto *p, byte restart, byte code)
 void
 proto_set_message(struct proto *p, char *msg, int len)
 {
-  PROTO_COMMON_LOCK(pcm);
-
   mb_free(p->message);
   p->message = NULL;
 
@@ -2411,7 +2288,7 @@ proto_set_message(struct proto *p, char *msg, int len)
   if (!len)
     return;
 
-  p->message = mb_alloc(pcm->pool, len + 1);
+  p->message = mb_alloc(proto_pool, len + 1);
   memcpy(p->message, msg, len);
   p->message[len] = 0;
 }
@@ -2718,20 +2595,20 @@ channel_show_stats(struct channel *c)
     cli_msg(-1006, "    Routes:         %u imported, %u exported, %u preferred",
 	    in_routes, out_routes, SRI(pref));
 
-  cli_msg(-1006, "    Route change stats:     received   rejected   filtered    ignored   RX limit   IN limit   accepted");
+  cli_msg(-1006, "    Route change stats:     received   rejected   filtered    ignored   RX limit      limit   accepted");
   cli_msg(-1006, "      Import updates:     %10u %10u %10u %10u %10u %10u %10u",
 	  SCI(updates_received), SCI(updates_invalid),
 	  SCI(updates_filtered), SRI(updates_ignored),
 	  SCI(updates_limited_rx), SCI(updates_limited_in),
 	  SRI(updates_accepted));
-  cli_msg(-1006, "      Import withdraws:   %10u %10u        --- %10u        --- %10u",
+  cli_msg(-1006, "      Import withdraws:   %10u %10u        --- %10u        ---        --- %10u",
 	  SCI(withdraws_received), SCI(withdraws_invalid),
 	  SRI(withdraws_ignored), SRI(withdraws_accepted));
-  cli_msg(-1006, "      Export updates:     %10u %10u %10u        --- %10u %10u",
-	  SRE(updates_received), SCE(updates_rejected),
-	  SCE(updates_filtered), SCE(updates_limited), SCE(updates_accepted));
-  cli_msg(-1006, "      Export withdraws:   %10u        ---        ---        ---         ---%10u",
-	  SRE(withdraws_received), SCE(withdraws_accepted));
+  cli_msg(-1006, "      Export updates:     %10u %10u %10u %10u        --- %10u %10u",
+	  SRE(updates_received), SCE(updates_rejected),	SCE(updates_filtered),
+	  SCE(updates_ignored), SCE(updates_limited), SCE(updates_accepted));
+  cli_msg(-1006, "      Export withdraws:   %10u        ---        --- %10u        ---        --- %10u",
+	  SRE(withdraws_received), SCE(withdraws_ignored), SCE(withdraws_accepted));
 
 #undef SRI
 #undef SRE
@@ -2872,11 +2749,7 @@ proto_cmd_enable(struct proto *p, uintptr_t arg, int cnt UNUSED)
   log(L_INFO "Enabling protocol %s", p->name);
   p->disabled = 0;
   proto_set_message(p, (char *) arg, -1);
-  struct proto_rethink_goal_deferred prgd = {
-    .dc.hook = proto_rethink_goal_deferred,
-    .p = p,
-  };
-  defer_call(&prgd.dc, sizeof prgd);
+  proto_rethink_goal(p);
   cli_msg(-11, "%s: enabled", p->name);
 }
 
