@@ -113,8 +113,9 @@ struct proto_config {
   u8 late_if_feed;			/* Delay interface feed after channels are up */
   u32 debug, mrtdump;			/* Debugging bitfields, both use D_* constants */
   u32 router_id;			/* Protocol specific router ID */
+  const char *hostname;			/* Protocol specific hostname */
   uint loop_order;			/* Launch a birdloop on this locking level; use DOMAIN_ORDER(the_bird) for mainloop */
-  btime loop_max_latency;		/* Request this specific maximum latency of loop; zero to default */
+  struct thread_group_config *thread_group;	/* Request this thread group for the birdloop */
   btime restart_limit;			/* Minimum allowed time between limit restarts */
 
   list channels;			/* List of channel configs (struct channel_config) */
@@ -302,8 +303,7 @@ struct proto_reload_request {
 
 #define PROTO_LEAVE_FROM_MAIN(loop) ({ if (loop != &main_birdloop) birdloop_leave(loop); })
 
-#define PROTO_LOCKED_FROM_MAIN(p)	for (struct birdloop *_proto_loop = PROTO_ENTER_FROM_MAIN(p); _proto_loop; \
-proto_announce_state((p), (p)->ea_state), PROTO_LEAVE_FROM_MAIN(_proto_loop), (_proto_loop = NULL))
+#define PROTO_LOCKED_FROM_MAIN(p)	for (struct birdloop *_proto_loop = PROTO_ENTER_FROM_MAIN(p); _proto_loop; PROTO_LEAVE_FROM_MAIN(_proto_loop), (_proto_loop = NULL))
 
 static inline struct domain_generic *proto_domain(struct proto *p)
 { return birdloop_domain(p->loop); }
@@ -314,6 +314,12 @@ proto_get_router_id(struct proto_config *pc)
   return pc->router_id ?: atomic_load_explicit(&global_runtime, memory_order_relaxed)->router_id;
 }
 
+static inline const char*
+proto_get_hostname(struct proto_config *pc)
+{
+  return pc->hostname ? pc->hostname : pc->global->hostname;
+}
+
 
 extern pool *proto_pool;
 
@@ -639,10 +645,12 @@ struct channel {
 
   struct channel_export_stats {
     /* Export - from core to protocol */
+    u32 updates_ignored;		/* Number of route updates ignored (squashed) by channel */
     u32 updates_rejected;		/* Number of route updates rejected by protocol */
     u32 updates_filtered;		/* Number of route updates rejected by filters */
     u32 updates_accepted;		/* Number of route updates accepted and exported */
     u32 updates_limited;		/* Number of route updates exceeding the out_limit */
+    u32 withdraws_ignored;		/* Number of route withdraws ignored (squashed) by channel */
     u32 withdraws_accepted;		/* Number of route withdraws accepted and processed */
   } export_stats;
 
@@ -660,12 +668,7 @@ struct channel {
   u8 stale;				/* Used in reconfiguration */
 
   u8 channel_state;
-  PACKED enum channel_reloadable {
-    CHANNEL_RELOADABLE_NEVER = 0,	/* Never reload */
-    CHANNEL_RELOADABLE_REMOTELY = 1,	/* Hook reload_routes() causes remote actions */
-    CHANNEL_RELOADABLE_LOCALLY = 2,	/* Hook reload_routes() reloads routes locally */
-  } reloadable;
-
+  u8 reloadable;			/* Hook reload_routes() is allowed on the channel */
   OBSREF(struct graceful_recovery_context) gr_lock;	/* Graceful restart mechanism should wait for this channel */
   u8 gr_wait;				/* Route export to channel is postponed until graceful restart */
 
@@ -691,7 +694,7 @@ struct channel {
  *
  * CS_DOWN - The initial and the final state of a channel. There is no route
  * exchange between the protocol and the table. Channel is not counted as
- * active. Channel keeps a ptr to the table, but do not lock the table and is
+ * active. Channel keeps a ptr to the table, keeps the table locked, but is
  * not linked in the table. Generally, new closed channels are created in
  * protocols' init() hooks. The protocol is expected to explicitly activate its
  * channels (by calling channel_init() or channel_open()).
