@@ -1058,6 +1058,7 @@ bgp_format_unknown(const eattr *a, byte *buf, uint size)
 static union bgp_attr_desc bgp_attr_table[BGP_ATTR_MAX] = {
   [BA_ORIGIN] = {
     .name = "bgp_origin",
+    .legacy_name = "BGP.origin",
     .type = T_ENUM_BGP_ORIGIN,
     .flags = BAF_TRANSITIVE,
     .export = bgp_export_origin,
@@ -1067,6 +1068,7 @@ static union bgp_attr_desc bgp_attr_table[BGP_ATTR_MAX] = {
   },
   [BA_AS_PATH] = {
     .name = "bgp_path",
+    .legacy_name = "BGP.as_path",
     .type = T_PATH,
     .flags = BAF_TRANSITIVE,
     .encode = bgp_encode_as_path,
@@ -1074,6 +1076,7 @@ static union bgp_attr_desc bgp_attr_table[BGP_ATTR_MAX] = {
   },
   [BA_NEXT_HOP] = {
     .name = "bgp_next_hop",
+    .legacy_name = "BGP.next_hop",
     .type = T_IP,
     .flags = BAF_TRANSITIVE,
     .encode = bgp_encode_next_hop,
@@ -1082,6 +1085,7 @@ static union bgp_attr_desc bgp_attr_table[BGP_ATTR_MAX] = {
   },
   [BA_MULTI_EXIT_DISC] = {
     .name = "bgp_med",
+    .legacy_name = "BGP.med",
     .type = T_INT,
     .flags = BAF_OPTIONAL,
     .encode = bgp_encode_u32,
@@ -1089,6 +1093,7 @@ static union bgp_attr_desc bgp_attr_table[BGP_ATTR_MAX] = {
   },
   [BA_LOCAL_PREF] = {
     .name = "bgp_local_pref",
+    .legacy_name = "BGP.local_pref",
     .type = T_INT,
     .flags = BAF_TRANSITIVE,
     .export = bgp_export_local_pref,
@@ -1097,6 +1102,7 @@ static union bgp_attr_desc bgp_attr_table[BGP_ATTR_MAX] = {
   },
   [BA_ATOMIC_AGGR] = {
     .name = "bgp_atomic_aggr",
+    .legacy_name = "BGP.atomic_aggr",
     .type = T_OPAQUE,
     .flags = BAF_TRANSITIVE,
     .encode = bgp_encode_raw,
@@ -1104,6 +1110,7 @@ static union bgp_attr_desc bgp_attr_table[BGP_ATTR_MAX] = {
   },
   [BA_AGGREGATOR] = {
     .name = "bgp_aggregator",
+    .legacy_name = "BGP.aggregator",
     .type = T_OPAQUE,
     .flags = BAF_OPTIONAL | BAF_TRANSITIVE,
     .encode = bgp_encode_aggregator,
@@ -1112,6 +1119,7 @@ static union bgp_attr_desc bgp_attr_table[BGP_ATTR_MAX] = {
   },
   [BA_COMMUNITY] = {
     .name = "bgp_community",
+    .legacy_name = "BGP.community",
     .type = T_CLIST,
     .flags = BAF_OPTIONAL | BAF_TRANSITIVE,
     .export = bgp_export_community,
@@ -1120,6 +1128,7 @@ static union bgp_attr_desc bgp_attr_table[BGP_ATTR_MAX] = {
   },
   [BA_ORIGINATOR_ID] = {
     .name = "bgp_originator_id",
+    .legacy_name = "BGP.originator_id",
     .type = T_QUAD,
     .flags = BAF_OPTIONAL,
     .export = bgp_export_originator_id,
@@ -1128,6 +1137,7 @@ static union bgp_attr_desc bgp_attr_table[BGP_ATTR_MAX] = {
   },
   [BA_CLUSTER_LIST] = {
     .name = "bgp_cluster_list",
+    .legacy_name = "BGP.cluster_list",
     .type = T_CLIST,
     .flags = BAF_OPTIONAL,
     .export = bgp_export_cluster_list,
@@ -1137,6 +1147,7 @@ static union bgp_attr_desc bgp_attr_table[BGP_ATTR_MAX] = {
   },
   [BA_MP_REACH_NLRI] = {
     .name = "bgp_mp_reach_nlri",
+    .legacy_name = "BGP.mp_reach_nlri",
     .type = T_OPAQUE,
     .hidden = 1,
     .flags = BAF_OPTIONAL,
@@ -1144,6 +1155,7 @@ static union bgp_attr_desc bgp_attr_table[BGP_ATTR_MAX] = {
   },
   [BA_MP_UNREACH_NLRI] = {
     .name = "bgp_mp_unreach_nlri",
+    .legacy_name = "BGP.mp_unreach_nlri",
     .type = T_OPAQUE,
     .hidden = 1,
     .flags = BAF_OPTIONAL,
@@ -1151,6 +1163,7 @@ static union bgp_attr_desc bgp_attr_table[BGP_ATTR_MAX] = {
   },
   [BA_EXT_COMMUNITY] = {
     .name = "bgp_ext_community",
+    .legacy_name = "BGP.ext_community",
     .type = T_ECLIST,
     .flags = BAF_OPTIONAL | BAF_TRANSITIVE,
     .export = bgp_export_ext_community,
@@ -1159,6 +1172,7 @@ static union bgp_attr_desc bgp_attr_table[BGP_ATTR_MAX] = {
   },
   [BA_AS4_PATH] = {
     .name = "bgp_as4_path",
+    .legacy_name = "BGP.as4_path",
     .type = T_PATH,
     .hidden = 1,
     .flags = BAF_OPTIONAL | BAF_TRANSITIVE,
@@ -1167,6 +1181,7 @@ static union bgp_attr_desc bgp_attr_table[BGP_ATTR_MAX] = {
   },
   [BA_AS4_AGGREGATOR] = {
     .name = "bgp_as4_aggregator",
+    .legacy_name = "BGP.as4_aggregator",
     .type = T_OPAQUE,
     .hidden = 1,
     .flags = BAF_OPTIONAL | BAF_TRANSITIVE,
@@ -1176,6 +1191,7 @@ static union bgp_attr_desc bgp_attr_table[BGP_ATTR_MAX] = {
   },
   [BA_AIGP] = {
     .name = "bgp_aigp",
+    .legacy_name = "BGP.aigp",
     .type = T_OPAQUE,
     .flags = BAF_OPTIONAL | BAF_DECODE_FLAGS,
     .export = bgp_export_aigp,
@@ -1185,6 +1201,7 @@ static union bgp_attr_desc bgp_attr_table[BGP_ATTR_MAX] = {
   },
   [BA_LARGE_COMMUNITY] = {
     .name = "bgp_large_community",
+    .legacy_name = "BGP.large_community",
     .type = T_LCLIST,
     .flags = BAF_OPTIONAL | BAF_TRANSITIVE,
     .export = bgp_export_large_community,
@@ -1193,6 +1210,7 @@ static union bgp_attr_desc bgp_attr_table[BGP_ATTR_MAX] = {
   },
   [BA_ONLY_TO_CUSTOMER] = {
     .name = "bgp_otc",
+    .legacy_name = "BGP.otc",
     .type = T_INT,
     .flags = BAF_OPTIONAL | BAF_TRANSITIVE,
     .encode = bgp_encode_u32,
@@ -1200,6 +1218,7 @@ static union bgp_attr_desc bgp_attr_table[BGP_ATTR_MAX] = {
   },
   [BA_MPLS_LABEL_STACK] = {
     .name = "bgp_mpls_label_stack",
+    .legacy_name = "BGP.mpls_label_stack",
     .type = T_CLIST,
     .readonly = 1,
     .export = bgp_export_mpls_label_stack,
@@ -1878,7 +1897,7 @@ bgp_get_prefix(struct bgp_ptx_private *c, struct netindex *ni, struct rte_src *s
 static void bgp_free_prefix(struct bgp_ptx_private *c, struct bgp_prefix *px);
 
 static inline int
-bgp_update_prefix(struct bgp_ptx_private *c, struct bgp_prefix *px, struct bgp_bucket *b, list *px_free_later)
+bgp_update_prefix(struct bgp_ptx_private *c, struct bgp_prefix *px, struct bgp_bucket *b)
 {
 #define IS_WITHDRAW_BUCKET(b)	((b) == c->withdraw_bucket)
 #define BPX_TRACE(what)	do { \
@@ -1911,11 +1930,7 @@ bgp_update_prefix(struct bgp_ptx_private *c, struct bgp_prefix *px, struct bgp_b
 
     /* Well, we haven't sent anything yet */
     if (!px->last)
-      if (px_free_later)
-	/* We may not be allowed to free the prefix directly */
-	add_tail(px_free_later, &px->buck_node);
-      else
-	bgp_free_prefix(c, px);
+      bgp_free_prefix(c, px);
 
     px->cur = NULL;
     return 0;
@@ -2000,7 +2015,6 @@ bgp_tx_resend(struct bgp_proto *p, struct bgp_channel *bc)
 
     ASSERT_DIE(bc->tx_keep);
 
-    list px_free_later; init_list(&px_free_later);
     HASH_WALK(c->prefix_hash, next, px)
     {
       if (!px->cur)
@@ -2013,7 +2027,7 @@ bgp_tx_resend(struct bgp_proto *p, struct bgp_channel *bc)
 	px->last = NULL;
 
 	/* And send it once again */
-	seen += bgp_update_prefix(c, px, last, &px_free_later);
+	seen += bgp_update_prefix(c, px, last);
       }
     }
     HASH_WALK_END;
@@ -2022,13 +2036,6 @@ bgp_tx_resend(struct bgp_proto *p, struct bgp_channel *bc)
       log(L_TRACE "%s.%s: TX resending %u routes",
 	  bc->c.proto->name, bc->c.name, seen);
 
-    /* Flush withdrawals */
-    struct bgp_prefix *px;
-    WALK_LIST_FIRST(px, px_free_later)
-    {
-      rem_node(&px->buck_node);
-      bgp_free_prefix(c, px);
-    }
   }
 
   if (p->enhanced_refresh)
@@ -2134,7 +2141,6 @@ bgp_init_pending_tx(struct bgp_channel *c)
   bpp->exporter = (struct rt_exporter) {
     .journal = {
       .loop = c->c.proto->loop,
-      .domain = dom.rtable,
       .item_size = sizeof(struct rt_export_item),
       .item_done = bgp_out_item_done,
     },
@@ -2170,19 +2176,12 @@ bgp_free_pending_tx(struct bgp_channel *bc)
 
   /* Move all prefixes to the withdraw bucket to unref the "last" prefixes */
   struct bgp_bucket *b = bgp_get_withdraw_bucket(c);
-  list px_free_later; init_list(&px_free_later);
   HASH_WALK(c->prefix_hash, next, px)
-    bgp_update_prefix(c, px, b, &px_free_later);
+    bgp_update_prefix(c, px, b);
   HASH_WALK_END;
 
   /* Flush withdrawals */
   struct bgp_prefix *px;
-  WALK_LIST_FIRST(px, px_free_later)
-  {
-    rem_node(&px->buck_node);
-    bgp_free_prefix(c, px);
-  }
-
   WALK_LIST_FIRST(px, b->prefixes)
     bgp_done_prefix(c, px, b);
 
@@ -2230,10 +2229,6 @@ bgp_preexport(struct channel *C, rte *e)
   struct bgp_proto *src = bgp_rte_proto(e);
   struct bgp_channel *c = (struct bgp_channel *) C;
 
-  /* Reject everything on shutdown */
-  if (SHUTTING_DOWN)
-    return -1;
-
   /* Ignore non-BGP channels */
   if (C->class != &channel_bgp)
     return -1;
@@ -2437,7 +2432,7 @@ bgp_update_attrs(struct bgp_proto *p, struct bgp_channel *c, rte *e, ea_list *at
   return bgp_export_attrs(&s, attrs);
 }
 
-void
+void NONNULL(3)
 bgp_rt_notify(struct proto *P, struct channel *C, const net_addr *n, rte *new, const rte *old)
 {
   struct bgp_proto *p = (void *) P;
@@ -2445,10 +2440,6 @@ bgp_rt_notify(struct proto *P, struct channel *C, const net_addr *n, rte *new, c
   struct bgp_bucket *buck;
   struct rte_src *path;
 
-  /* Reject everything on shutdown */
-  if (SHUTTING_DOWN)
-    return;
-
   /* Ignore non-BGP channels */
   if (C->class != &channel_bgp)
     return;
@@ -2466,7 +2457,7 @@ bgp_rt_notify(struct proto *P, struct channel *C, const net_addr *n, rte *new, c
   path = (new ?: old)->src;
 
   /* And queue the notification */
-  if (bgp_update_prefix(c, bgp_get_prefix(c, NET_TO_INDEX(n), path, bc->add_path_tx), buck, NULL))
+  if (bgp_update_prefix(c, bgp_get_prefix(c, NET_TO_INDEX(n), path, bc->add_path_tx), buck))
     bgp_schedule_packet(p->conn, bc, PKT_UPDATE);
 
   if ((p->cf->tx_size_warning > 0) )
@@ -2523,7 +2514,7 @@ bgp_rte_better(const rte *new, const rte *old)
   if (n < o)
     return 0;
 
-  /* LLGR draft - depreference stale routes */
+  /* LLGR 9494 - depreference stale routes */
   n = rte_stale(new);
   o = rte_stale(old);
   if (n > o)
@@ -2660,7 +2651,7 @@ bgp_rte_mergable(const rte *pri, const rte *sec)
   if (rte_resolvable(pri) != rte_resolvable(sec))
     return 0;
 
-  /* LLGR draft - depreference stale routes */
+  /* LLGR 9494 - depreference stale routes */
   if (rte_stale(pri) != rte_stale(sec))
     return 0;
 
